
#include "GeomIntr2DLinLin.h"
using namespace Geom;

//----------------------------------------------------------------------------
static bool Find (const Vector2& _P0, const Vector2& _D0,
    const Vector2& _P1, const Vector2& _D1, Vector2& _Diff,
    Real& rfD0SqrLen, int& riQuantity, Real afT[2])
{
    // Intersection is a solution to P0+s*D0 = P1+t*D1.  Rewrite as
    // s*D0 - t*D1 = P1 - P0, a 2x2 system of equations.  If D0 = (x0,y0)
    // and D1 = (x1,y1) and P1 - P0 = (c0,c1), then the system is
    // x0*s - x1*t = c0 and y0*s - y1*t = c1.  The error tests are relative
    // to the size of the direction vectors, |Cross(D0,D1)| >= e*|D0|*|D1|
    // rather than absolute tests |Cross(D0,D1)| >= e.  The quantities
    // P1-P0, |D0|^2, and |D1|^2 are returned for use by calling functions.

    Real fDet = _D1.x*_D0.y - _D1.y*_D0.x;
    _Diff = _P1 - _P0;
    rfD0SqrLen = _D0.SquaredLength();

    const Real fEpsilon = 1e-06f;
    if ( fDet*fDet > fEpsilon*rfD0SqrLen*_D1.SquaredLength() )
    {
        // Lines intersect in a single point.  Return both s and t values for
        // use by calling functions.
        Real fInvDet = 1.0f/fDet;
        riQuantity = 1;
        afT[0] = (_D1.x*_Diff.y - _D1.y*_Diff.x)*fInvDet;
        afT[1] = (_D0.x*_Diff.y - _D0.y*_Diff.x)*fInvDet;
    }
    else
    {
        // lines are parallel
        fDet = _D0.x*_Diff.y - _D0.y*_Diff.x;
        if ( fDet*fDet > fEpsilon*rfD0SqrLen*_Diff.SquaredLength() )
        {
            // lines are disjoint
            riQuantity = 0;
        }
        else
        {
            // lines are the same
            riQuantity = 2;
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Line2& _Line0, const Line2& _Line1,
    int& riQuantity, Real afT[2])
{
    Vector2 kDiff;
    Real fD0SqrLen;
    bool bIntersects = Find(_Line0.Origin(),_Line0.Direction(),
        _Line1.Origin(),_Line1.Direction(),kDiff,fD0SqrLen,riQuantity,afT);

    if ( bIntersects )
    {
        if ( riQuantity == 2 )
        {
            afT[0] = -Math::MAX_REAL;
            afT[1] = Math::MAX_REAL;
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Line2& _Line, const Ray2& _Ray,
    int& riQuantity, Real afT[2])
{
    Vector2 kDiff;
    Real fD0SqrLen;
    bool bIntersects = Find(_Line.Origin(),_Line.Direction(),
        _Ray.Origin(),_Ray.Direction(),kDiff,fD0SqrLen,riQuantity,afT);

    if ( bIntersects )
    {
        if ( riQuantity == 1 )
        {
            if ( afT[1] < 0.0f )
            {
                // lines intersect, but ray does not intersect line
                riQuantity = 0;
            }
        }
        else
        {
            // ray is contained by line, adjust intersection interval
            if ( _Line.Direction().Dot(_Ray.Direction()) > 0.0f )
            {
                afT[0] = (kDiff.Dot(_Line.Direction()))/fD0SqrLen;
                afT[1] = Math::MAX_REAL;
            }
            else
            {
                afT[0] = -Math::MAX_REAL;
                afT[1] = (kDiff.Dot(_Line.Direction()))/fD0SqrLen;
            }
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Line2& _Line, const Segment2& _Segment,
    int& riQuantity, Real afT[2])
{
    Vector2 kDiff;
    Real fD0SqrLen;
    bool bIntersects = Find(_Line.Origin(),_Line.Direction(),
        _Segment.Origin(),_Segment.Direction(),kDiff,fD0SqrLen,riQuantity,
        afT);

    if ( bIntersects )
    {
        if ( riQuantity == 1 )
        {
            if ( afT[1] < 0.0f || afT[1] > 1.0f )
            {
                // lines intersect, but segment does not intersect line
                riQuantity = 0;
            }
        }
        else
        {
            // segment is contained by line, adjust intersection interval
            Real fDot = _Line.Direction().Dot(_Segment.Direction());
            Real fInvLen = 1.0f/fD0SqrLen;
            if ( fDot > 0.0f )
            {
                afT[0] = (kDiff.Dot(_Line.Direction()))*fInvLen;
                afT[1] = afT[0] + fDot*fInvLen;
            }
            else
            {
                afT[1] = (kDiff.Dot(_Line.Direction()))*fInvLen;
                afT[0] = afT[1] + fDot*fInvLen;
            }
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Ray2& _Ray0, const Ray2& _Ray1,
    int& riQuantity, Real afT[2])
{
    Vector2 kDiff;
    Real fD0SqrLen;
    bool bIntersects = Find(_Ray0.Origin(),_Ray0.Direction(),
        _Ray1.Origin(),_Ray1.Direction(),kDiff,fD0SqrLen,riQuantity,afT);

    if ( bIntersects )
    {
        if ( riQuantity == 1 )
        {
            if ( afT[0] < 0.0f || afT[1] < 0.0f )
            {
                // lines intersect, but rays do not
                riQuantity = 0;
            }
        }
        else
        {
            // rays are on same line
            Real fDot = kDiff.Dot(_Ray0.Direction());
            if ( _Ray0.Direction().Dot(_Ray1.Direction()) > 0.0f )
            {
                // Rays point in same direction, get semiinfinite interval
                // of intersection.
                afT[0] = ( fDot >= 0.0f ? fDot/fD0SqrLen : 0.0f );
                afT[1] = Math::MAX_REAL;
            }
            else
            {
                // test for overlap of opposite pointing rays
                if ( fDot > 0.0f )
                {
                    afT[0] = 0.0f;
                    afT[1] = fDot/fD0SqrLen;
                }
                else if ( fDot < 0.0f )
                {
                    // no overlap
                    riQuantity = 0;
                }
                else
                {
                    // rays have a common origin
                    riQuantity = 1;
                    afT[0] = 0.0f;
                }
            }
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Ray2& _Ray, const Segment2& _Segment,
    int& riQuantity, Real afT[2])
{
    Vector2 kDiff;
    Real fD0SqrLen;
    bool bIntersects = Find(_Ray.Origin(),_Ray.Direction(),
        _Segment.Origin(),_Segment.Direction(),kDiff,fD0SqrLen,riQuantity,
        afT);

    if ( bIntersects )
    {
        if ( riQuantity == 1 )
        {
            if ( afT[0] < 0.0f || afT[1] < 0.0f || afT[1] > 1.0f )
            {
                // lines intersect, but ray and segment do not
                riQuantity = 0;
            }
        }
        else
        {
            // ray and segment are on the same line
            Real fDotRS = _Ray.Direction().Dot(_Segment.Direction());
            Real fDot0, fDot1;
            if ( fDotRS > 0.0f )
            {
                fDot0 = kDiff.Dot(_Ray.Direction());
                fDot1 = fDot0 + fDotRS;
            }
            else
            {
                fDot1 = kDiff.Dot(_Ray.Direction());
                fDot0 = fDot1 + fDotRS;
            }

            // compute intersection of [t0,t1] and [0,+infinity]
            if ( fDot0 >= 0.0f )
            {
                // complete overlap
                Real fInvLen = 1.0f/fD0SqrLen;
                afT[0] = fDot0*fInvLen;
                afT[1] = fDot1*fInvLen;
            }
            else if ( fDot1 > 0.0f )
            {
                // partial overlap
                afT[0] = 0.0f;
                afT[1] = fDot1/fD0SqrLen;
            }
            else if ( fDot1 < 0.0f )
            {
                // no overlap
                riQuantity = 0;
            }
            else
            {
                // overlap at a single end point
                riQuantity = 1;
                afT[0] = 0.0f;
            }
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------
bool Geom::FindIntersection (const Segment2& _Segment0,
    const Segment2& _Segment1, int& riQuantity, Real afT[2])
{
    Vector2 kDiff;
    Real fD0SqrLen;
    bool bIntersects = Find(_Segment0.Origin(),_Segment0.Direction(),
        _Segment1.Origin(),_Segment1.Direction(),kDiff,fD0SqrLen,
        riQuantity,afT);

    if ( bIntersects )
    {
        if ( riQuantity == 1 )
        {
            if ( afT[0] < 0.0f || afT[0] > 1.0f
            ||   afT[1] < 0.0f || afT[1] > 1.0f )
            {
                // lines intersect, but segments do not
                riQuantity = 0;
            }
        }
        else
        {
            // segments are on the same line
            Real fDotRS =
                _Segment0.Direction().Dot(_Segment1.Direction());
            Real fDot0, fDot1;
            if ( fDotRS > 0.0f )
            {
                fDot0 = kDiff.Dot(_Segment0.Direction());
                fDot1 = fDot0 + fDotRS;
            }
            else
            {
                fDot1 = kDiff.Dot(_Segment0.Direction());
                fDot0 = fDot1 + fDotRS;
            }

            // compute intersection of [t0,t1] and [0,1]
            if ( fDot1 < 0.0f || fDot0 > fD0SqrLen )
            {
                riQuantity = 0;
            }
            else if ( fDot1 > 0.0f )
            {
                if ( fDot0 < fD0SqrLen )
                {
                    Real fInvLen = 1.0f/fD0SqrLen;
                    riQuantity = 2;
                    afT[0] = ( fDot0 < 0.0f ? 0.0f : fDot0*fInvLen );
                    afT[1] = ( fDot1 > fD0SqrLen ? 1.0f : fDot1*fInvLen );
                }
                else  // fT0 == 1
                {
                    riQuantity = 1;
                    afT[0] = 1.0f;
                }
            }
            else  // fT1 == 0
            {
                riQuantity = 1;
                afT[0] = 0.0f;
            }
        }
    }

    return riQuantity != 0;
}
//----------------------------------------------------------------------------


